#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifndef DEBIAN
#include <linux/ip.h>
#include <linux/icmp.h>
#else
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#endif
extern int errno;

#ifndef NOFILE
#define NOFILE 1024
#endif

#ifndef DEBIAN
int     go_background(void);
#else
void    go_background(void);
#endif
char   *hostlookup(unsigned long int);

#ifdef DEBIAN
#include "filelog.c"
#endif

struct ippkt {
	struct iphdr ip;
	struct icmphdr icmp;
	char    buffer[10000];
} pkt;

#ifndef DEBIAN
int     go_background(void)
#else
void    go_background(void)
#endif
{
	int     fd;
	int     fs;

	if (getppid() != 1) {
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
		fs = fork();
		if (fs < 0) {
			perror("fork");
			exit(1);
		}
		if (fs > 0)
			exit(0);
#ifndef DEBIAN
		setpgrp();
#endif
		fd = open("/dev/tty", O_RDWR);
		if (fd >= 0) {
			ioctl(fd, TIOCNOTTY, (char *) NULL);
			close(fd);
		}
	}
	for (fd = 0; fd < NOFILE; fd++)
		close(fd);
	errno = 0;
	chdir("/");
	umask(0);
#ifdef DEBIAN
	setsid();
#endif
}

int     main(int argc,
    char *argv[])
{
	int     s;
#ifndef DEBIAN
	int     i;
#else
	struct logger log;
#endif
	int		icmp_dest_unreach = 1;
	int		icmp_source_quench = 1;
	int		icmp_redirect = 1;
	int		icmp_echo = 1;
	int		icmp_info_request = 1;

	setuid(0);
	if (geteuid() != 0) {
		printf("This program requires root privledges\n");
		exit(0);
	}
#ifndef DEBIAN
	while (*argv) {
		if (!strcmp(*argv, "-d")) // backward compability
			icmp_dest_unreach = 0;
		else if (!strcmp(*argv, "-qd"))
			icmp_dest_unreach = 0;
		else if (!strcmp(*argv, "-qs"))
			icmp_source_quench = 0;
		else if (!strcmp(*argv, "-qr"))
			icmp_redirect = 0;
		else if (!strcmp(*argv, "-qe"))
			icmp_echo = 0;
		else if (!strcmp(*argv, "-qi"))
			icmp_info_request = 0;
		else
			fprintf(stderr, "ERROR: unknown option %s\n", *argv);
		argv++;
	}
#else
	argv++;
	log.file = 0;
	log.f = (void *) syslog;
	while(*argv) {
		if (!strcmp(*argv, "-d")) // backward compability
			icmp_dest_unreach = 0;

		if (!strcmp(*argv, "-qd"))
			icmp_dest_unreach = 0;
		if (!strcmp(*argv, "-qs"))
			icmp_source_quench = 0;
		if (!strcmp(*argv, "-qr"))
			icmp_redirect = 0;
		if (!strcmp(*argv, "-qe"))
			icmp_echo = 0;
		if (!strcmp(*argv, "-qi"))
			icmp_info_request = 0;

		if (!strcmp(*argv,"-f")) {
			log.file = 1;
			log.f = (void *) filelog;
		}

		if (!strcmp(*argv,"--help")) {
			printf("icmplogd 1.1 [-d] [-f]: ICMP session logger\n\t-d  do not log ICMP destination unreachable packets\n\t-f  logs into a file rather than using syslog\n");
			return 0;
		}
		argv++;
	}
#endif
	go_background();
	s = socket(AF_INET, SOCK_RAW, 1);
#ifdef DEBIAN
	if (log.file == 0)
#endif
	openlog("icmplogd", 0, LOG_DAEMON);

#ifdef DEBIAN
#define syslog log.f
#endif
	while (1) {
		read(s, (struct ippkt *) &pkt, 9999);
		if (pkt.ip.ihl != 5) {
			syslog(LOG_NOTICE, "suspicious ip options from %s", hostlookup(pkt.ip.daddr));
			continue;
		}
		if (pkt.icmp.type == ICMP_DEST_UNREACH && icmp_dest_unreach)
			syslog(LOG_NOTICE, "destination unreachable from %s", hostlookup(pkt.ip.saddr));
		if (pkt.icmp.type == ICMP_SOURCE_QUENCH && icmp_source_quench)
			syslog(LOG_NOTICE, "source quench from %s", hostlookup(pkt.ip.saddr));
		if (pkt.icmp.type == ICMP_REDIRECT && icmp_redirect)
			syslog(LOG_NOTICE, "source route from %s", hostlookup(pkt.ip.saddr));
		if (pkt.icmp.type == ICMP_ECHO && icmp_echo)
			syslog(LOG_NOTICE, "ping from %s", hostlookup(pkt.ip.saddr));
		if (pkt.icmp.type == ICMP_INFO_REQUEST && icmp_info_request)
			syslog(LOG_NOTICE, "info request from %s", hostlookup(pkt.ip.saddr));
	}
#ifdef DEBIAN
#undef syslog
#endif
}

char   *hostlookup(unsigned long int in)
{
	static char blah[1024];
#ifdef DEBIAN
	static char address[24];
#endif
	struct in_addr i;
	struct hostent *he;

	i.s_addr = in;
	he = gethostbyaddr((char *) &i, sizeof(struct in_addr), AF_INET);

#ifndef DEBIAN
	if (he == NULL)
		strcpy(blah, inet_ntoa(i));
	else
		strcpy(blah, he->h_name);
#endif
	if(he) {
		strncpy(blah, he->h_name, 1000);
		blah[1000] = '\0';
		sprintf(address, " [%.20s]", inet_ntoa(i));
		strcat(blah, address);
	}
	else
		sprintf(blah, "[%.20s]", inet_ntoa(i));
	return blah;
}
