#include "tcplog.h"
#ifndef NOFILE
#define NOFILE 1024
#endif

#ifndef DEBIAN
int     go_background(void);
#else
void    go_background(void);
#endif
char   *hostlookup(__u32);
char   *servlookup(unsigned short);
extern char *get_rfc1413_data(int, int, __u32, char *);

#ifdef DEBIAN
#include "filelog.c"
#endif

struct ippkt {
	struct iphdr ip;
	struct tcphdr tcp;
	char    buffer[10000];
};

#ifndef DEBIAN
int     main(void)
#else
int	main(int argc, char **argv)
#endif
{
	int     s;
#ifndef DEBIAN
	int     i, status;
	char    tmpbuff[1024];
	int     forkstat;
#else
	int     status;
#endif
	char    remote_addr[1024];
#ifndef DEBIAN
	struct ippkt pkt, opkt;
#else
	struct ippkt pkt;
	struct logger log;
#endif

	setuid(0);
	if (geteuid() != 0) {
		printf("This program requires root privledges\n");
		exit(0);
	}
#ifdef DEBIAN
	argv++;
	log.file = 0;
	log.f = (void *) syslog;
	while(*argv) {
		if (!strcmp(*argv,"-f")) {
			log.file = 1;
			log.f = (void *) filelog;
		}
		if (!strcmp(*argv,"--help")) {
			printf("tcplogd 1.1 [-f]: TCP session logger\n\t-f  logs into a file rather than using syslog\n");
			return 0;
		}
		argv++;
	}
#endif
	go_background();
	s = socket(AF_INET, SOCK_RAW, 6);

	while (1) {
		read(s, (struct ippkt *) &pkt, 9999);
		if (pkt.tcp.syn == 1 && pkt.tcp.ack == 0) {
			if (!fork()) { /* double fork()    */
				if (!fork()) {	/* to avoid zombies */
#ifdef DEBIAN
#define syslog log.f

					if (log.file == 0)
#endif
					openlog("tcplogd", 0, LOG_DAEMON);
					strcpy(remote_addr, hostlookup(pkt.ip.saddr));
					if (ntohs(pkt.tcp.dest) != 113)
						ident_user = get_rfc1413_data(ntohs(pkt.tcp.source), ntohs(pkt.tcp.dest), pkt.ip.saddr, hostlookup(pkt.ip.daddr));
					else
						ident_user = remote_addr;

					if (ntohs(pkt.tcp.source) == 20 && ntohs(pkt.tcp.dest) < 1024)
						syslog(LOG_NOTICE, "FTPBounce attack detected from %s", ident_user);
					syslog(LOG_NOTICE, "%s connection attempt from %s", servlookup(pkt.tcp.dest), ident_user);
#ifdef DEBIAN
					if (ntohs(pkt.tcp.dest) != 113)
#endif
					free(ident_user);
#ifdef DEBIAN
#undef syslog
					if (log.file == 0)
#endif
					closelog();
					exit(0);
				}
				exit(0);
			} else
				wait(&status);
		}
	}
}

#ifndef DEBIAN
int     go_background(void)
#else
void    go_background(void)
#endif
{
	int     fd;
	int     fs;

	if (getppid() != 1) {
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
		fs = fork();
		if (fs < 0) {
			perror("fork");
			exit(1);
		}
		if (fs > 0)
			exit(0);
#ifndef DEBIAN
		setpgrp();
#endif
		fd = open("/dev/tty", O_RDWR);
		if (fd >= 0) {
			ioctl(fd, TIOCNOTTY, (char *) NULL);
			close(fd);
		}
	}
	for (fd = 0; fd < NOFILE; fd++)
		close(fd);
	errno = 0;
	chdir("/");
	umask(0);
#ifdef DEBIAN
	setsid();
#endif
}

char   *
        hostlookup(__u32 in)
{
	static char blah[1024];
#ifdef DEBIAN
	static char address[24];
#endif
	struct in_addr i;
	struct hostent *he;

	i.s_addr = in;
	he = gethostbyaddr((char *) &i, sizeof(struct in_addr), AF_INET);

#ifndef DEBIAN
	if (he == NULL)
		strcpy(blah, inet_ntoa(i));
	else
		strcpy(blah, he->h_name);
#endif
	if(he) {
		strncpy(blah, he->h_name, 1000);
		blah[1000] = '\0';
		sprintf(address, " [%.20s]", inet_ntoa(i));
		strcat(blah, address);
	}
	else
		sprintf(blah, "[%.20s]", inet_ntoa(i));
	return blah;
}

char   *
        servlookup(unsigned short port)
{
	struct servent *se;
	static char buff[1024];

	se = getservbyport(port, "tcp");
	if (se == NULL)
		sprintf(buff, "port %d", ntohs(port));
	else
		sprintf(buff, "%s", se->s_name);
	return buff;
}
